#!/bin/sh
# -------------------------------------------------------------------------
# (C) Copyright Marxmeier Software AG 2008-2014
#
# Eloquence JDLG Plugin Interface
#
# Function: Create CA-signed JAR certificate.
# -------------------------------------------------------------------------

# Configuration
# -------------

# Edit the values below as necessary.
#
# Notes:
# - TITLE and ORGANIZATION must follow X.500 Distinguished Names syntax.
#   Comma should be avoided or must be backslash-escaped.
# - ALIAS must follow file naming conventions. Avoid spaces and special
#   characters.
# - CA_ALIAS must be set as configured in create-ca.sh.
# - TITLE should match desc.title in build.xml.
# - ALIAS should match plugin.filename in build.xml.
# - ORGANIZATION should match desc.organization in build.xml.

TITLE="Eloquence JDLG Plugin"
ALIAS="jdlg-plugin"
ORGANIZATION="Marxmeier Software AG"
COUNTRY="DE"
DAYS_VALID=365
CA_ALIAS="jdlg-ca"

# Commands
OPENSSL="openssl"
KEYTOOL="keytool"

# -----------------
# End Configuration

if [ -n "$1" ]; then
  if [ "$1" = "-help" -o $# -ne 1 ]; then
    echo "usage: $0 [ALIAS]"
    exit 3
  fi
  ALIAS="$1"
fi

DNAME="C=$COUNTRY,O=$ORGANIZATION,CN=$TITLE"
CA_DIR="$CA_ALIAS"
SSL_CONF="$CA_DIR/openssl.cnf"
KEYSTORE="keystore.bin"
PASSWORDS_CONF="passwords.cnf"

cd "`dirname $0`"

if [ ! -d "$CA_DIR" -o ! -f "$SSL_CONF" -o ! -f "$KEYSTORE" ]; then
  echo "CA not configured. Use create-sa.sh first."
  exit 2
fi

if [ ! -r "$PASSWORDS_CONF" ]; then
  echo "Password configuration file not found."
  exit 2
fi
. "./$PASSWORDS_CONF"
if [ -z "$KEYSTORE_PASS" -o -z "$KEY_PASS" -o -z "$PRIVATE_KEY_PASS" ]; then
  echo "KEYSTORE_PASS or KEY_PASS or PRIVATE_KEY_PASS not configured."
  exit 2
fi

# Delete existing key.

"$KEYTOOL" -delete \
  -alias "$ALIAS" \
  -keystore "$KEYSTORE" \
  -storepass "$KEYSTORE_PASS" \
> /dev/null

# Generate key pair.

"$KEYTOOL" -genkeypair \
  -alias "$ALIAS" \
  -dname "$DNAME" \
  -validity "$DAYS_VALID" \
  -keyalg RSA \
  -keysize 2048 \
  -keystore "$KEYSTORE" \
  -storepass "$KEYSTORE_PASS" \
  -keypass "$KEY_PASS" \
|| exit 1

# Generate certificate request.

"$KEYTOOL" -certreq \
  -alias "$ALIAS" \
  -keyalg RSA \
  -keystore "$KEYSTORE" \
  -storepass "$KEYSTORE_PASS" \
  -keypass "$KEY_PASS" \
  -file "$ALIAS.req.p10" \
|| exit 1

# Generate signed certificate.

"$OPENSSL" ca \
  -batch -config "$SSL_CONF" \
  -days "$DAYS_VALID" \
  -passin "pass:$PRIVATE_KEY_PASS" \
  -out "$ALIAS.cert.pem" \
  -infiles "$ALIAS.req.p10" \
|| exit 1

rm "$ALIAS.req.p10"

# Import certificate,
# requires to be DER-formatted.

"$OPENSSL" x509 \
  -in "$ALIAS.cert.pem" \
  -out "$ALIAS.cert.der" \
  -outform DER \
|| exit 1

rm "$ALIAS.cert.pem"

"$KEYTOOL" -importcert \
  -v -trustcacerts \
  -alias "$ALIAS" \
  -keystore "$KEYSTORE" \
  -storepass "$KEYSTORE_PASS" \
  -keypass "$KEY_PASS" \
  -file "$ALIAS.cert.der" \
|| exit 1

rm "$ALIAS.cert.der"

echo
echo "CA-signed JAR certificate for $ALIAS successfully created in $KEYSTORE."
echo
