/* ------------------------------------------------------------------------
   (C) Copyright Marxmeier Software AG 2009-2015

   Project:  Eloquence JDLG Plugin Interface
   File:     Editgrid.java
   Function: JDLG Plugin: Editable grid
   $Id: Editgrid.java,v 1.33 2015/02/19 18:13:46 rhg Exp $

   Edit history:
   2009-04-30, ap   - creation
   2009-11-26, ap   - finalized
   2014-04-08, rhg  - api version 1.1 additions
   2014-12-04, rhg  - fixed onClone() issues
                    - editingStopped() fixed swapped cx/cy
                    - recognize keypad cursor keys
   2015-02-19, rhg  - api version 1.2: handle right mouse button
 ------------------------------------------------------------------------ */

package com.eloquence.plugin;

import com.eloquence.api.*;

import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.table.*;
import javax.swing.border.*;
import javax.swing.event.*;

public class Editgrid
   extends JScrollPane
   implements Plugin, FocusListener, KeyListener,
              TableColumnModelListener, CellEditorListener
{
   private Jdlg jdlg = null;
   private Vector columnWidths = new Vector();
   private JList rowHeader = new JList(new DefaultListModel());
   private Vector readonlyColumns = new Vector();
   private JLabel corner = new JLabel();
   private CellRenderer cellRenderer = new CellRenderer();
   private ColumnHeaderRenderer headerRenderer = new ColumnHeaderRenderer();
   private RowHeaderRenderer rowRenderer = new RowHeaderRenderer();

   private static Color uiFocusCellForegroundColor;
   private static Color uiFocusCellBackgroundColor;
   private static Border uiCellBorder;
   private static Border uiFocusCellBorder;

   public Editgrid()
   {
      super(new JTable());

      if (uiFocusCellForegroundColor == null)
         uiFocusCellForegroundColor = UIManager.getColor(
            "TableHeader.focusCellForeground");
      if (uiFocusCellBackgroundColor == null)
         uiFocusCellBackgroundColor = UIManager.getColor(
            "TableHeader.focusCellBackground");
      if (uiCellBorder == null)
         uiCellBorder = UIManager.getBorder(
            "TableHeader.cellBorder");
      if (uiFocusCellBorder == null)
         uiFocusCellBorder = UIManager.getBorder(
            "TableHeader.focusCellBorder");

      JTable table = (JTable)getViewport().getView();
      table.setModel(new EditGridModel(0, 0));
      JTableHeader header = table.getTableHeader();
      header.setReorderingAllowed(false);
      header.setResizingAllowed(false);
      table.getColumnModel().addColumnModelListener(this);

      corner.setHorizontalAlignment(JLabel.LEFT);
      corner.setBorder(uiCellBorder);
      corner.setOpaque(true);
      setCorner(JScrollPane.UPPER_LEFT_CORNER, corner);

      rowHeader.setCellRenderer(rowRenderer);

      setAutoscrolls(true);
      setRowHeaderView(rowHeader);
   }

   // ------------------------------------------------------------------------
   // Plugin
   // ------------------------------------------------------------------------

   public Component onCreate(final Jdlg jdlg)
   {
      if (!jdlg.requiresApiVersion(1, 0))
         return null;

      this.jdlg = jdlg;

      final JTable table = (JTable)getViewport().getView();
      DefaultCellEditor cellEditor =
         (DefaultCellEditor)table.getDefaultEditor(String.class);
      final JTextField cellEditorTextField =
         (JTextField)cellEditor.getComponent();

      // Install mouse listener on table, rowHeader and corner
      // to handle right mouse button (requires API version >= 1.2).

      MouseAdapter mouseListener = new MouseAdapter()
      {
         public void mousePressed(MouseEvent event)
         {
            if (jdlg.requiresApiVersion(1, 2))
            {
               // API version >= 1.2:
               // Handle right mouse button pressed event,
               // eg. open the context menu.
               jdlg.processRightMouseButtonEvent(event);
            }
         }
      };

      table.addMouseListener(mouseListener);
      rowHeader.addMouseListener(mouseListener);
      corner.addMouseListener(mouseListener);

      // Install mouse listener on cell editor:
      // - obtain focus
      // - make caret visible
      // - handle right mouse button (requires API version >= 1.2).

      cellEditorTextField.addMouseListener(
         new MouseAdapter()
         {
            public void mousePressed(MouseEvent event)
            {
               if (jdlg.requiresApiVersion(1, 2))
               {
                  // API version >= 1.2:
                  // Handle right mouse button pressed event,
                  // eg. open the context menu.
                  if (jdlg.processRightMouseButtonEvent(event))
                     return;
               }

               if (!table.hasFocus())
                  jdlg.obtainFocus();
            }

            public void mouseReleased(MouseEvent event)
            {
               if (jdlg.requiresApiVersion(1, 2))
               {
                  // API version >= 1.2:
                  // Consume right mouse button released event.
                  if (jdlg.consumeRightMouseButtonEvent(event))
                     return;
               }

               cellEditorTextField.getCaret().setVisible(true);
            }
         }
      );

      cellEditor.setClickCountToStart(0);
      cellEditor.addCellEditorListener(this);
      table.setFocusTraversalKeysEnabled(false);
      table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
      table.addKeyListener(this);
      table.addFocusListener(this);
      cellEditorTextField.addFocusListener(this);
      table.setSurrendersFocusOnKeystroke(true);

      return this;
   }

   public void onDispose()
   {
   }

   public void onEloq2Swing()
   {
      if (cxCyModified)
      {
         selectCell(cy, cx);
         jdlg.obtainFocus();
         cxCyModified = false;
      }

      if (headerModified)
      {
         rescaleHeader();
         headerModified = false;
      }
   }


   public void onSwing2Eloq()
   {
      if (cxSelected >= 0 && cySelected >= 0)
      {
         cx = cxSelected;
         cy = cySelected;
         cxCyModified = false;
      }
   }

   public void onFontModified(Font font)
   {
      int height = getFontMetrics(font).getHeight();

      corner.setFont(font);
      rowHeader.setFixedCellHeight(height);

      JTable table = (JTable)getViewport().getView();
      DefaultCellEditor cellEditor =
         (DefaultCellEditor)table.getDefaultEditor(String.class);
      Component cellEditorComponent = cellEditor.getComponent();

      table.setRowHeight(height);
      cellEditorComponent.setFont(font);

      rescaleHeader();
      setRowHeaderWidth();
   }

   public boolean isFocusable()
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      return tableModel.getRowCount() > 0;
   }

   public void onFocusReceived()
   {
   }

   public void onFocusLost()
   {
   }

   public Component onGetFocusPeer()
   {
      return getViewport().getView();
   }

   public void onTabFocus(boolean toNext)
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      int rowCount = tableModel.getRowCount(),
          columnCount = tableModel.getColumnCount();
      if (rowCount < 1 || columnCount < 1)
         return;

      if (toNext)
      {
         cx = 0;
         cy = 0;
      }
      else {
         cx = columnCount - 1;
         cy = rowCount - 1;
      }

      cxCyModified = true;

      table.setColumnSelectionInterval(cx, cx);
      table.setRowSelectionInterval(cy, cy);
      scrollToVisible(cy, cx);
   }

   public boolean fireRuleOnLosingFocus()
   {
      return false;
   }

   public boolean onRuleOverride()
   {
      return false;
   }

   public boolean allowRMBR()
   {
      return true;
   }

   // ------------------------------------------------------------------------
   // FocusListener
   // ------------------------------------------------------------------------

   public void focusGained(FocusEvent e)
   {
   }

   public void focusLost(FocusEvent e)
   {
      JTable table = (JTable)getViewport().getView();
      if (table == e.getOppositeComponent())
         return;

      DefaultCellEditor cellEditor =
         (DefaultCellEditor)table.getDefaultEditor(String.class);
      Component cellEditorComponent = cellEditor.getComponent();
      if (cellEditorComponent == e.getOppositeComponent())
         return;

      cellEditor.cancelCellEditing();
   }

   // ------------------------------------------------------------------------
   // KeyListener
   // ------------------------------------------------------------------------

   private boolean forceUpdate = false;

   public void keyPressed(KeyEvent event)
   {
      JTable table;
      switch (event.getKeyCode())
      {
         case KeyEvent.VK_TAB:
            table = (JTable)getViewport().getView();
            if (event.isShiftDown())
            {
               if (   table.getSelectedColumn() == 0
                   && table.getSelectedRow() == 0)
	          break;
            }
            else {
               if (   table.getSelectedColumn() == table.getColumnCount() -1
                   && table.getSelectedRow() == table.getRowCount() -1)
	          break;
            }
            return;

         case KeyEvent.VK_LEFT:
         case KeyEvent.VK_KP_LEFT:
            table = (JTable)getViewport().getView();
            if (table.getSelectedColumn() == 0 && table.getRowCount() > 1)
            {
               int row = table.getSelectedRow();
               if (row > 0)
               {
                  cy = row - 1;
                  cx = table.getColumnCount() - 1;
                  forceUpdate = true;
                  return;
               }
            }
            break;

         case KeyEvent.VK_RIGHT:
         case KeyEvent.VK_KP_RIGHT:
            table = (JTable)getViewport().getView();
            if (table.getSelectedColumn() == table.getColumnCount() - 1)
            {
               int rows = table.getRowCount();
               if (rows > 1)
               {
                  int row = table.getSelectedRow();
                  if (row < rows - 1)
                  {
                     cy = row + 1;
                     cx = 0;
                     forceUpdate = true;
                     return;
                  }
               }
            }
            break;

         case KeyEvent.VK_UP:
         case KeyEvent.VK_KP_UP:
            table = (JTable)getViewport().getView();
            if (table.getSelectedRow() > 0)
               return;
            break;

         case KeyEvent.VK_DOWN:
         case KeyEvent.VK_KP_DOWN:
            table = (JTable)getViewport().getView();
            if (table.getSelectedRow() < table.getRowCount() - 1)
               return;
            break;
      }

      jdlg.getKeyboardListener().keyPressed(event);
   }

   public void keyTyped(KeyEvent event)
   {
      jdlg.getKeyboardListener().keyTyped(event);
   }

   public void keyReleased(KeyEvent event)
   {
      if (forceUpdate)
      {
         selectCell(cy, cx);
         forceUpdate = false;
      }

      jdlg.getKeyboardListener().keyReleased(event);
   }

   // ------------------------------------------------------------------------
   // TableColumnModelListener
   // ------------------------------------------------------------------------

   public void columnAdded(TableColumnModelEvent e)
   {
   }

   public void columnMarginChanged(ChangeEvent e)
   {
      if (!headerModified)
      {
         JTable table = (JTable)getViewport().getView();
         DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
         DefaultTableColumnModel columnModel =
            (DefaultTableColumnModel)table.getColumnModel();

         Dimension grid = jdlg.getGridDimension();
         int columnCount = tableModel.getColumnCount();

         for (int column = 0; column < columnCount; column++)
         {
            int width = (int)Math.ceil(
               (double)columnModel.getColumn(column).getWidth() / grid.width);
            columnWidths.set(column, new Integer(width));
         }
      }
   }

   public void columnMoved(TableColumnModelEvent e)
   {
   }

   public void columnRemoved(TableColumnModelEvent e)
   {
   }

   public void columnSelectionChanged(ListSelectionEvent e)
   {
      JTable table = (JTable)getViewport().getView();
      cySelected = table.getSelectedRow();
      cxSelected = table.getSelectedColumn();
      if (cxSelected < 0 || cySelected < 0)
         cxSelected = cySelected = -1;
   }

   // ------------------------------------------------------------------------
   // CellEditorListener
   // ------------------------------------------------------------------------

   public void editingCanceled(ChangeEvent e)
   {
   }

   public void editingStopped(ChangeEvent e)
   {
      JTable table = (JTable)getViewport().getView();
      cySelected = table.getSelectedRow();
      cxSelected = table.getSelectedColumn();
      jdlg.triggerRule();
   }

   // ------------------------------------------------------------------------
   // Events
   // ------------------------------------------------------------------------

   public void processEvent(AWTEvent event)
   {
      if (!jdlg.ignoreEvent(event))
         super.processEvent(event);
   }

   private void selectCell(int row, int column)
   {
      cxSelected = cySelected = -1;

      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      if (row < 0 || row >= tableModel.getRowCount())
         return;
      if (column < 0 || column >= tableModel.getColumnCount())
         return;

      table.setRowSelectionInterval(row, row);
      table.setColumnSelectionInterval(column, column);
      scrollToVisible(row, column);
   }

   private void scrollToVisible(int row, int column)
   {
      JViewport viewport = getViewport();
      JTable table = (JTable)viewport.getView();
      Rectangle rect = table.getCellRect(row, column, true);
      Point pt = viewport.getViewPosition();
      rect.setLocation(rect.x - pt.x, rect.y - pt.y);
      viewport.scrollRectToVisible(rect);
   }

   // ------------------------------------------------------------------------
   // Table header
   // ------------------------------------------------------------------------

   private int fixedRowHeaderWidth = -1;

   private void setRowHeaderWidth()
   {
      Dimension grid = jdlg.getGridDimension();
      int maxWidth = 0;

      if (fixedRowHeaderWidth < 0)
      {
         for(int row = 0; row < rowHeader.getModel().getSize(); row ++)
         {
            Rectangle2D rowRect = getFontMetrics(getFont()).getStringBounds(
               (String)rowHeader.getModel().getElementAt(row),
               getGraphics());

            if (rowRect.getWidth() > maxWidth)
               maxWidth = (int) rowRect.getWidth();
         }

         maxWidth = (int)Math.ceil((double)maxWidth / grid.width) * grid.width;
      }
      else
         maxWidth = fixedRowHeaderWidth * grid.width;

      rowHeader.setFixedCellWidth(maxWidth + 2);
   }

   private void rescaleHeader()
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      DefaultTableColumnModel columnModel =
         (DefaultTableColumnModel)table.getColumnModel();

      Dimension grid = jdlg.getGridDimension();
      int columnCount = tableModel.getColumnCount();

      for (int column = 0; column < columnCount; column++)
      {
         int width = ((Integer)columnWidths.get(column)).intValue();
         columnModel.getColumn(column).setPreferredWidth(width * grid.width);
      }
   }

   // ------------------------------------------------------------------------
   // Foreground/background colors
   // ------------------------------------------------------------------------

   public void setForeground(Color color)
   {
      super.setForeground(color);
      if (rowHeader != null)
         rowHeader.setForeground(color);
   }

   public void setBackground(Color color)
   {
      super.setBackground(color);
      if (rowHeader != null)
         rowHeader.setBackground(color);

      JViewport viewport = getViewport();
      if (viewport != null)
      {
         viewport.setBackground(color);
         viewport.invalidate();

         Component view = viewport.getView();
         if (view != null)
            view.setBackground(color);
      }
   }

   // ------------------------------------------------------------------------
   // Parse cell values from line
   // ------------------------------------------------------------------------

   private final Vector parseLine(String line)
   {
      StringTokenizer cellTokenizer = new StringTokenizer(line, delim, true);
      Vector parsed = new Vector();
      String lastToken = delim;

      while (cellTokenizer.hasMoreTokens())
      {
         String token = cellTokenizer.nextToken();
         if (!token.equals(delim))
         {
            parsed.add(token);
            if (!cellTokenizer.hasMoreTokens())
               break;

            token = cellTokenizer.nextToken();
         }
         else if (lastToken.equals(delim))
            parsed.add("");

         lastToken = token;
      }

      return parsed;
   }

   // ------------------------------------------------------------------------
   // get/set content/line/cell
   // ------------------------------------------------------------------------

   private static final String CONTENT_ATTR_NAME = "content";
   private static final String LINE_ATTR_NAME = "line";
   private static final String CELL_ATTR_NAME = "cell";

   private boolean inContent = false;

   private final void setContent(String content)
   {
      setClear();
      inContent = true;

      StringTokenizer lineTokenizer = new StringTokenizer(content, "\n");
      while (lineTokenizer.hasMoreTokens())
         setAdd(lineTokenizer.nextToken());

      inContent = false;
      setRowHeaderWidth();
   }

   private boolean inLine = false;

   private final boolean setLine(int row, String rowContent)
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      if (row < 1 || row > tableModel.getRowCount())
         return false;

      inLine = true;

      Vector parsed = parseLine(rowContent);
      int columnCount = Math.min(parsed.size(), getColumns());
      for (int column = 0; column < columnCount; column++)
         if (!setCell(row - 1, column, (String)parsed.get(column)))
            return false;

      inLine = false;
      if (!inContent)
         setRowHeaderWidth();
      return true;
   }

   private final boolean setCell(int row, int column, String cellContent)
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      if (row < 1 || row > tableModel.getRowCount())
         return false;
      if (column < 0 || column > tableModel.getColumnCount())
         return false;

      if (column == 0)
      {
         DefaultListModel headerModel = (DefaultListModel)rowHeader.getModel();
         headerModel.set(row - 1, cellContent);
      }
      else {
         tableModel.setValueAt(cellContent, row - 1, column - 1);
         if (!inLine && !inContent)
            setRowHeaderWidth();
      }

      return true;
   }

   private final String getContent()
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();

      String content = "";

      int rowCount = tableModel.getRowCount();
      for (int row = 1; row <= rowCount; row++)
         content += getLine(row) + "\n";

      return content;
   }

   private final String getLine(int row)
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      if (row < 1 || row > tableModel.getRowCount())
         return null;

      String content = "";

      int columnCount = tableModel.getColumnCount();
      for (int column = 0; column <= columnCount; column++)
         content += getCell(row, column) + delim;

      return content;
   }

   private final String getCell(int row, int column)
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      if (row < 1 || row > tableModel.getRowCount())
         return null;
      if (column < 0 || column > tableModel.getColumnCount())
         return null;

      if (column == 0)
      {
         DefaultListModel headerModel = (DefaultListModel)rowHeader.getModel();
         return (String)headerModel.get(row - 1);
      }

      return (String)tableModel.getValueAt(row - 1, column - 1);
   }

   // ------------------------------------------------------------------------
   // get/set columns
   // ------------------------------------------------------------------------

   private static final String COLUMNS_ATTR_NAME = "columns";

   private final boolean setColumns(int columns)
   {
      if (columns < 1)
         return false;

      JTable table = (JTable)getViewport().getView();
      table.setModel(new EditGridModel(0, columns));

      DefaultTableColumnModel columnModel =
         (DefaultTableColumnModel)table.getColumnModel();

      JTableHeader header = table.getTableHeader();
      DefaultTableColumnModel headerColumnModel =
         (DefaultTableColumnModel)header.getColumnModel();

      readonlyColumns = new Vector();
      columnWidths = new Vector();

      for (int i = 0; i < columns; i++)
      {
         headerColumnModel.getColumn(i).setHeaderValue("");

         TableColumn column = columnModel.getColumn(i);
         column.setCellRenderer(cellRenderer);
         column.setHeaderRenderer(headerRenderer);

         readonlyColumns.add(new Boolean(false));
         columnWidths.add(new Integer(5));
      }

      rescaleHeader();
      return true;
   }

   private final int getColumns()
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      return tableModel.getColumnCount();
   }

   // ------------------------------------------------------------------------
   // get/set column
   // ------------------------------------------------------------------------

   private static final String HEADER_ATTR_NAME = "header";

   private final boolean setHeader(int column, String columnName)
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();

      if (column < 0 || column > tableModel.getColumnCount())
         return false;

      if (column == 0)
         corner.setText(columnName);
      else {
         JTableHeader header = table.getTableHeader();
         DefaultTableColumnModel headerColumnModel =
            (DefaultTableColumnModel)header.getColumnModel();
         headerColumnModel.getColumn(column - 1).setHeaderValue(columnName);
      }

      return true;
   }

   private final String getHeader(int column)
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();

      if (column < 0 || column > tableModel.getColumnCount())
         return null;

      if (column == 0)
         return corner.getText();

      JTableHeader header = table.getTableHeader();
      DefaultTableColumnModel headerColumnModel =
         (DefaultTableColumnModel)header.getColumnModel();
      return (String)headerColumnModel.getColumn(column - 1).getHeaderValue();
   }

   // ------------------------------------------------------------------------
   // set add
   // ------------------------------------------------------------------------

   private static final String ADD_ATTR_NAME = "add";

   private final boolean setAdd(String add)
   {
      Vector parsed = parseLine(add);

      DefaultListModel headerModel = (DefaultListModel)rowHeader.getModel();
      headerModel.addElement(parsed.firstElement());

      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      tableModel.addRow(parsed.subList(1, parsed.size()).toArray());

      setRowHeaderWidth();
      return true;
   }

   // ------------------------------------------------------------------------
   // set insert
   // ------------------------------------------------------------------------

   private static final String INSERT_ATTR_NAME = "insert";

   private final boolean setInsert(int row, String insert)
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      if (row < 1 || row > tableModel.getRowCount())
         return false;

      Vector parsed = parseLine(insert);

      DefaultListModel headerModel = (DefaultListModel)rowHeader.getModel();
      headerModel.insertElementAt(parsed.firstElement(), row - 1);
      tableModel.insertRow(row - 1, parsed.subList(1, parsed.size()).toArray());

      setRowHeaderWidth();
      return true;
   }

   // ------------------------------------------------------------------------
   // set del
   // ------------------------------------------------------------------------

   private static final String DEL_ATTR_NAME = "del";

   private final boolean setDel(int row, int lines)
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      if (row < 1 || row > tableModel.getRowCount())
         return false;

      DefaultListModel headerModel = (DefaultListModel)rowHeader.getModel();
      headerModel.remove(row - 1);
      tableModel.removeRow(row - 1);

      setRowHeaderWidth();
      return true;
   }

   // ------------------------------------------------------------------------
   // get/set delim
   // ------------------------------------------------------------------------

   private static final String DELIM_ATTR_NAME = "delim";
   protected String delim = "\t";

   private final void setDelim(String delim)
   {
      this.delim = delim;
   }

   private final String getDelim()
   {
      return delim;
   }

   // ------------------------------------------------------------------------
   // get/set colwidth
   // ------------------------------------------------------------------------

   private static final String COLWIDTH_ATTR_NAME = "colwidth";
   private boolean headerModified = true;

   private final boolean setColWidth(int column, int width)
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      if (column < 0 || column > tableModel.getColumnCount())
         return false;

      if (column == 0)
         fixedRowHeaderWidth = width;
      else {
         if (width < 0)
            width = 0;

         columnWidths.set(column - 1, new Integer(width));
         headerModified = true;
      }

      return true;
   }

   private final int getColWidth(int column)
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      if (column < 0 || column > tableModel.getColumnCount())
         return -1;

      if (column == 0)
         return fixedRowHeaderWidth;
      return ((Integer)columnWidths.get(column - 1)).intValue();
   }

   // ------------------------------------------------------------------------
   // get/set colreadonly
   // ------------------------------------------------------------------------

   private static final String COLREADONLY_ATTR_NAME = "colreadonly";

   private final boolean setColReadOnly(int column, int readOnly)
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      if (column < 1 || column > tableModel.getColumnCount())
         return false;

      readonlyColumns.set(column - 1, new Boolean(readOnly != 0));
      return true;
   }

   private final int getColReadOnly(int column)
   {
      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      if (column < 1 || column > tableModel.getColumnCount())
         return -1;

      return ((Boolean)readonlyColumns.get(column - 1)).booleanValue() ? 1 : 0;
   }

   // ------------------------------------------------------------------------
   // get/set cx/activecolumn
   // ------------------------------------------------------------------------

   private static final String CX_ATTR_NAME = "cx";
   private static final String ACTIVECOLUMN_ATTR_NAME = "activecolumn";
   private boolean cxCyModified = false;
   private int cx = -1;
   private int cxSelected = -1;

   private final boolean setCx(int cx)
   {
      this.cx = cx - 1;
      cxCyModified = true;
      return true;
   }

   private final int getCx()
   {
      return cx + 1;
   }

   // ------------------------------------------------------------------------
   // get/set cy/activeline
   // ------------------------------------------------------------------------

   private static final String CY_ATTR_NAME = "cy";
   private static final String ACTIVELINE_ATTR_NAME = "activeline";
   private int cy = -1;
   private int cySelected = -1;

   private final boolean setCy(int cy)
   {
      this.cy = cy - 1;
      cxCyModified = true;
      return true;
   }

   private final int getCy()
   {
      return cy + 1;
   }

   // ------------------------------------------------------------------------
   // set clear
   // ------------------------------------------------------------------------

   private static final String CLEAR_ATTR_NAME = "clear";

   private final void setClear()
   {
      DefaultListModel headerModel = (DefaultListModel)rowHeader.getModel();
      headerModel.clear();

      JTable table = (JTable)getViewport().getView();
      DefaultTableModel tableModel = (DefaultTableModel)table.getModel();
      tableModel.setRowCount(0);

      cy = -1;
      cx = -1;
   }

   // ------------------------------------------------------------------------
   // attributes
   // ------------------------------------------------------------------------

   private static final int CONTENT_ATTR_ID = 1;
   private static final int LINE_ATTR_ID = 2;
   private static final int CELL_ATTR_ID = 3;
   private static final int COLUMNS_ATTR_ID = 4;
   private static final int HEADER_ATTR_ID = 5;
   private static final int CLEAR_ATTR_ID = 6;
   private static final int ADD_ATTR_ID = 7;
   private static final int INSERT_ATTR_ID = 8;
   private static final int DEL_ATTR_ID = 9;
   private static final int DELIM_ATTR_ID = 10;
   private static final int COLWIDTH_ATTR_ID = 11;
   private static final int COLREADONLY_ATTR_ID = 12;
   private static final int CX_ATTR_ID = 13;
   private static final int CY_ATTR_ID = 14;

   public void onClone(Plugin clone)
   {
      Editgrid cloneObj = (Editgrid)clone;
      cloneObj.setDelim(getDelim());

      int columnCount = getColumns();
      cloneObj.setColumns(columnCount);

      for (int column = 0; column <= columnCount; column++)
      {
         cloneObj.setHeader(column, getHeader(column));
         cloneObj.setColWidth(column, getColWidth(column));
         cloneObj.setColReadOnly(column, getColReadOnly(column));
      }

      cloneObj.setContent(getContent());
      cloneObj.setCx(getCx());
      cloneObj.setCy(getCy());
   }

   public boolean onSet(String attribute, java.util.List index,
                        UnifiedData data)
   {
      Integer attrInt = (Integer)attributeList.get(attribute);
      if (attrInt != null)
      {
         int indexElements = index != null ? index.size() : 0;
         switch (attrInt.intValue())
         {
            case CONTENT_ATTR_ID:
               if (indexElements != 0)
	          return false;
               if (!data.isStringData())
	          return false;
               setContent(data.getStringData());
               break;

            case LINE_ATTR_ID:
            {
               if (indexElements != 1)
	          return false;
               if (!data.isStringData())
	          return false;
               UnifiedData rowIndex = (UnifiedData)index.get(0);
               if (rowIndex.isStringData())
	          return false;
               setLine(rowIndex.getIntData(), data.getStringData());
               break;
            }

            case CELL_ATTR_ID:
            {
               if (indexElements != 2)
	          return false;
               if (!data.isStringData())
	          return false;
               UnifiedData rowIndex = (UnifiedData)index.get(0);
               UnifiedData columnIndex = (UnifiedData)index.get(1);
               if (rowIndex.isStringData() || columnIndex.isStringData())
	          return false;
               setCell(rowIndex.getIntData(), columnIndex.getIntData(),
                       data.getStringData());
               break;
            }

            case COLUMNS_ATTR_ID:
               if (indexElements != 0)
	          return false;
               if (data.isStringData())
	          return false;
               return setColumns(data.getIntData());

            case HEADER_ATTR_ID:
            {
               if (indexElements != 1)
	          return false;
               if (!data.isStringData())
	          return false;
               UnifiedData columnIndex = (UnifiedData)index.get(0);
               if (columnIndex.isStringData())
	          return false;
               return setHeader(columnIndex.getIntData(), data.getStringData());
            }

            case CLEAR_ATTR_ID:
               if (indexElements != 0)
	          return false;
               setClear();
               break;

            case ADD_ATTR_ID:
               if (indexElements != 0)
	          return false;
               if (!data.isStringData())
	          return false;
               return setAdd(data.getStringData());

            case INSERT_ATTR_ID:
            {
               if (indexElements != 1)
	          return false;
               if (!data.isStringData())
	          return false;
               UnifiedData rowIndex = (UnifiedData)index.get(0);
               if (rowIndex.isStringData())
	          return false;
               return setInsert(rowIndex.getIntData(), data.getStringData());
            }

            case DEL_ATTR_ID:
            {
               if (indexElements != 1)
	          return false;
               if (data.isStringData())
	          return false;
               UnifiedData rowIndex = (UnifiedData)index.get(0);
               if (rowIndex.isStringData())
	          return false;
               setDel(rowIndex.getIntData(), data.getIntData());
               break;
            }

            case DELIM_ATTR_ID:
               if (indexElements != 0)
	          return false;
               if (!data.isStringData())
	          return false;
               setDelim(data.getStringData());
               break;

            case COLWIDTH_ATTR_ID:
            {
               if (indexElements != 1)
	          return false;
               if (data.isStringData())
	          return false;
               UnifiedData columnIndex = (UnifiedData)index.get(0);
               if (columnIndex.isStringData())
	          return false;
               setColWidth(columnIndex.getIntData(), data.getIntData());
               break;
            }

            case COLREADONLY_ATTR_ID:
            {
               if (indexElements != 1)
	          return false;
               if (data.isStringData())
	          return false;
               UnifiedData columnIndex = (UnifiedData)index.get(0);
               if (columnIndex.isStringData())
	          return false;
               setColReadOnly(columnIndex.getIntData(), data.getIntData());
               break;
            }

            case CX_ATTR_ID:
               if (indexElements != 0)
	          return false;
               if (data.isStringData())
	          return false;
               return setCx(data.getIntData());

            case CY_ATTR_ID:
               if (indexElements != 0)
	          return false;
               if (data.isStringData())
	          return false;
               return setCy(data.getIntData());
         }

         return true;
      }

      return false;
   }

   public UnifiedData onGet(String attribute, java.util.List index)
   {
      Integer attrInt = (Integer)attributeList.get(attribute);
      if (attrInt != null)
      {
         int indexElements = index != null ? index.size() : 0;
         switch(attrInt.intValue())
         {
            case CONTENT_ATTR_ID:
               if (indexElements != 0)
	          return null;
               return new UnifiedData(getContent());

            case LINE_ATTR_ID:
            {
               if (indexElements != 1)
	          return null;
               UnifiedData rowIndex = (UnifiedData)index.get(0);
               if (rowIndex.isStringData())
	          return null;
               return new UnifiedData(getLine(rowIndex.getIntData()));
            }

            case CELL_ATTR_ID:
            {
               if (indexElements != 2)
	          return null;
               UnifiedData rowIndex = (UnifiedData)index.get(0);
               UnifiedData columnIndex = (UnifiedData)index.get(1);
               if (rowIndex.isStringData() || columnIndex.isStringData())
	          return null;
               return new UnifiedData(getCell(rowIndex.getIntData(),
                                              columnIndex.getIntData()));
            }

            case COLUMNS_ATTR_ID:
               if (indexElements != 0)
	          return null;
               return new UnifiedData(getColumns());

            case HEADER_ATTR_ID:
            {
               if (indexElements != 1)
	          return null;
               UnifiedData columnIndex = (UnifiedData)index.get(0);
               if (columnIndex.isStringData())
	          return null;
               return new UnifiedData(getHeader(columnIndex.getIntData()));
            }

            case DELIM_ATTR_ID:
               if (indexElements != 0)
	          return null;
               return new UnifiedData(getDelim());

            case COLWIDTH_ATTR_ID:
            {
               if (indexElements != 1)
	          return null;
               UnifiedData columnIndex = (UnifiedData)index.get(0);
               if (columnIndex.isStringData())
	          return null;
               int colWidth = getColWidth(columnIndex.getIntData());
               if (colWidth < 0)
	          return null;
               return new UnifiedData(colWidth);
            }

            case COLREADONLY_ATTR_ID:
            {
               if (indexElements != 1)
	          return null;
               UnifiedData columnIndex = (UnifiedData)index.get(0);
               if (columnIndex.isStringData())
	          return null;
               int colReadOnly = getColReadOnly(columnIndex.getIntData());
               if (colReadOnly < 0)
	          return null;
               return new UnifiedData(colReadOnly);
            }

            case CX_ATTR_ID:
               if (indexElements != 0)
	          return null;
               return new UnifiedData(getCx());

            case CY_ATTR_ID:
               if (indexElements != 0)
	          return null;
               return new UnifiedData(getCy());
         }
      }

      return null;
   }

   private static HashMap attributeList = new HashMap();
   static
   {
      attributeList.put(CONTENT_ATTR_NAME, new Integer(CONTENT_ATTR_ID));
      attributeList.put(LINE_ATTR_NAME, new Integer(LINE_ATTR_ID));
      attributeList.put(CELL_ATTR_NAME, new Integer(CELL_ATTR_ID));
      attributeList.put(COLUMNS_ATTR_NAME, new Integer(COLUMNS_ATTR_ID));
      attributeList.put(HEADER_ATTR_NAME, new Integer(HEADER_ATTR_ID));
      attributeList.put(CLEAR_ATTR_NAME, new Integer(CLEAR_ATTR_ID));
      attributeList.put(ADD_ATTR_NAME, new Integer(ADD_ATTR_ID));
      attributeList.put(INSERT_ATTR_NAME, new Integer(INSERT_ATTR_ID));
      attributeList.put(DEL_ATTR_NAME, new Integer(DEL_ATTR_ID));
      attributeList.put(DELIM_ATTR_NAME, new Integer(DELIM_ATTR_ID));
      attributeList.put(COLWIDTH_ATTR_NAME, new Integer(COLWIDTH_ATTR_ID));
      attributeList.put(COLREADONLY_ATTR_NAME, new Integer(COLREADONLY_ATTR_ID));
      attributeList.put(CX_ATTR_NAME, new Integer(CX_ATTR_ID));
      attributeList.put(ACTIVECOLUMN_ATTR_NAME, new Integer(CX_ATTR_ID));
      attributeList.put(CY_ATTR_NAME, new Integer(CY_ATTR_ID));
      attributeList.put(ACTIVELINE_ATTR_NAME, new Integer(CY_ATTR_ID));
   }

   // ------------------------------------------------------------------------
   // Data model
   // ------------------------------------------------------------------------

   private class EditGridModel extends DefaultTableModel
   {
      public EditGridModel(int rows, int columns)
      {
         super(rows, columns);
      }

      public boolean isCellEditable(int row, int column)
      {
         return !((Boolean)readonlyColumns.get(column)).booleanValue();
      }
   }

   // ------------------------------------------------------------------------
   // Cell renderer
   // ------------------------------------------------------------------------

   private class CellRenderer extends DefaultTableCellRenderer
   {
      public Component getTableCellRendererComponent(
         JTable table, Object value,
         boolean isSelected, boolean hasFocus, int row, int column)
      {
         Component defComp = super.getTableCellRendererComponent(
            table, value, isSelected, hasFocus, row, column);

         defComp.setFont(Editgrid.this.getFont());
         defComp.setForeground(Editgrid.this.getForeground());
         defComp.setBackground(Editgrid.this.getBackground());
         return defComp;
      }
   }

   // ------------------------------------------------------------------------
   // Column header renderer
   // ------------------------------------------------------------------------

   private class ColumnHeaderRenderer extends DefaultTableCellRenderer
   {
      public Component getTableCellRendererComponent(
         JTable table, Object value,
         boolean isSelected, boolean hasFocus, int row, int column)
      {
         Component defComp = super.getTableCellRendererComponent(
            table, value, isSelected, hasFocus, row, column);

         defComp.setFont(Editgrid.this.getFont());

         if (table != null)
         {
            JTableHeader header = table.getTableHeader();
            if (header != null)
            {
               Color fgColor = null;
               Color bgColor = null;

               if (hasFocus)
               {
                  fgColor = uiFocusCellForegroundColor;
                  bgColor = uiFocusCellBackgroundColor;
               }

               if (fgColor == null)
                  fgColor = header.getForeground();

               if (bgColor == null)
                  bgColor = header.getBackground();

               setForeground(fgColor);
               setBackground(bgColor);
            }

            setText(value == null ? "" : value.toString());

            Border border = null;

            if (hasFocus)
               border = uiFocusCellBorder;
            if (border == null)
               border = uiCellBorder;
            setBorder(border);
         }

         return defComp;
      }
   }

   // ------------------------------------------------------------------------
   // Row header renderer
   // ------------------------------------------------------------------------

   private class RowHeaderRenderer extends DefaultListCellRenderer
   {
      public Component getListCellRendererComponent(
         JList list, Object value, int index,
         boolean isSelected, boolean cellHasFocus)
      {
         Component defComp = super.getListCellRendererComponent(
            list, value, index, isSelected, cellHasFocus);

         defComp.setFont(Editgrid.this.getFont());
         defComp.setBackground(Editgrid.this.getBackground());
         return defComp;
      }

      public void paintComponent(Graphics g)
      {
         super.paintComponent(g);
         setBorder(uiCellBorder);
      }
   }
}
