/* ------------------------------------------------------------------------
   (C) Copyright Marxmeier Software AG 2009-2014

   Project:  Eloquence JDLG Plugin Interface
   File:     Pie.java
   Function: JDLG Charts Plugin: Pie Chart
   $Id: Pie.java,v 1.8 2014/12/11 12:17:23 rhg Exp $

   Edit history:
   2009-04-06, ap   - creation
   2009-04-28, mike - finalized
   2014-04-08, rhg  - api version 1.1 additions
   2014-12-03, rhg  - revised, enhancements, bugfixes, code cleanup
 ------------------------------------------------------------------------ */

package com.eloquence.plugin.charts;

import com.eloquence.api.*;
import com.eloquence.services.tools.*;

import java.awt.*;
import java.util.*;
import java.text.*;

import org.jfree.chart.*;
import org.jfree.chart.plot.*;
import org.jfree.data.general.*;
import org.jfree.chart.labels.*;
import org.jfree.util.*;
import org.jfree.ui.*;

public class Pie
   extends Chart
{
   private ChartPanel peerPanel;
   private DefaultPieDataset dataset = new DefaultPieDataset();
   private boolean datasetModified = true;

   // ------------------------------------------------------------------------
   // Plugin
   // ------------------------------------------------------------------------

   public Component onCreate(Jdlg jdlg)
   {
      if (!jdlg.requiresApiVersion(1, 0))
         return null;

      peerPanel = new ChartPanel(null);
      return peerPanel;
   }

   public void onDispose()
   {
      peerPanel = null;
      dataset = null;
   }

   public void onEloq2Swing()
   {
      String titleLabel = null;

      if (chartModified)
      {
         // createPieChart(java.lang.String title,
         //                PieDataset dataset,
         //                boolean legend,
         //                boolean tooltips,
         //                boolean urls)

         if (titleLabel == null)
            titleLabel = title.length() > 0 ? title : null;

         if (paint3d)
            peerPanel.setChart(
               ChartFactory.createPieChart3D(
                  titleLabel, dataset,
                  legend, true, false));
         else
            peerPanel.setChart(
               ChartFactory.createPieChart(
                  titleLabel, dataset,
                  legend, true, false));

         chartModified = false;
         attributeModified = true;
         datasetModified = true;
      }

      JFreeChart chart = null;
      PiePlot plot = null;

      if (datasetModified)
      {
         if (chart == null)
         {
            chart = peerPanel.getChart();
            plot = (PiePlot)chart.getPlot();
         }

         dataset.clear();

         Iterator it = content.entrySet().iterator();
	 while (it.hasNext())
	 {
            Map.Entry e = (Map.Entry)it.next();
            String label = (String)e.getKey();
            Section section = (Section)e.getValue();

            dataset.setValue(label, section.value);
            if (section.color != null)
               plot.setSectionPaint(label, section.color);
         }

         dataset.sortByKeys(SortOrder.ASCENDING);
      }

      if (attributeModified)
      {
         if (chart == null)
         {
            chart = peerPanel.getChart();
            plot = (PiePlot)chart.getPlot();
         }

         if (titleLabel == null)
            titleLabel = title.length() > 0 ? title : null;
         chart.setTitle(titleLabel);

         chart.setBorderVisible(border);
         chart.setPadding(borderInsets);
         plot.setBackgroundPaint(chartBgc);
         plot.setOutlineVisible(outline);
         plot.setForegroundAlpha((float)fgAlpha / 100);

         if (plot instanceof PiePlot3D)
            ((PiePlot3D)plot).setDarkerSides(true);

         PieSectionLabelGenerator labelGenerator;
         if (showLabels)
            labelGenerator = new StandardPieSectionLabelGenerator(
               getLabelFormat(), valueFormat(), percentFormat());
         else
            labelGenerator = null;
         plot.setLabelGenerator(labelGenerator);
      }

      if (datasetModified || attributeModified)
      {
         boolean explodeSet = explode.length() > 0;
         int itemCount = dataset.getItemCount();
         for (int section = 0; section < itemCount; section++)
         {
            Comparable key = dataset.getKey(section);

            double percent;
            if (explodeSet && key.compareTo(explode) == 0)
               percent = 0.3;
            else
               percent = 0;

            plot.setExplodePercent(key, percent);
         }
      }

      datasetModified = false;
      attributeModified = false;
   }

   // ------------------------------------------------------------------------
   // default labelformat
   // ------------------------------------------------------------------------

   protected String getDefaultLabelFormat()
   {
      return "{0} {2}";
   }

   // ------------------------------------------------------------------------
   // get/set paint3d
   // ------------------------------------------------------------------------

   private static final String PAINT3D_ATTR_NAME = "paint3d";
   private boolean paint3d = false;

   private final void setPaint3d(int paint3d)
   {
      this.paint3d = paint3d != 0;
      chartModified = true;
   }

   private final int getPaint3d()
   {
      return paint3d ? 1 : 0;
   }

   // ------------------------------------------------------------------------
   // get/set outline
   // ------------------------------------------------------------------------

   private static final String OUTLINE_ATTR_NAME = "outline";
   private boolean outline = false;

   private final void setOutline(int outline)
   {
      this.outline = outline != 0;
      attributeModified = true;
   }

   private final int getOutline()
   {
      return outline ? 1 : 0;
   }

   // ------------------------------------------------------------------------
   // get/set explode
   // ------------------------------------------------------------------------

   private static final String EXPLODE_ATTR_NAME = "explode";
   private String explode = "";

   private final void setExplode(String explode)
   {
      this.explode = explode;
      attributeModified = true;
   }

   private final String getExplode()
   {
      return explode;
   }

   // ------------------------------------------------------------------------
   // set clear
   // ------------------------------------------------------------------------

   private static final String CLEAR_ATTR_NAME = "clear";

   private final void setClear()
   {
      content.clear();
      datasetModified = true;
   }

   // ------------------------------------------------------------------------
   // get/set content
   // ------------------------------------------------------------------------

   private static final String CONTENT_ATTR_NAME = "content";
   private HashMap content = new LinkedHashMap();

   private final boolean parseSection(String section)
   {
      StringTokenizer cellTokenizer = new StringTokenizer(section, delim);
      if (cellTokenizer.hasMoreTokens())
      {
         String label = cellTokenizer.nextToken();
         if (!cellTokenizer.hasMoreTokens())
            return false;

         double value;
         try
         {
            value = new Double(cellTokenizer.nextToken()).doubleValue();
            if (value <= 0)
               return false;
         }
         catch (NumberFormatException e)
         {
            return false;
         }

         Color color = null;
         if (cellTokenizer.hasMoreTokens())
         {
            String colorString = cellTokenizer.nextToken().trim();
            int idx = colorString.indexOf('#');
            if (idx == -1)
               return false;

            try
            {
               color = new Color(Integer.parseInt(
                  colorString.substring(idx + 1), 16));
            }
            catch (NumberFormatException e)
            {
               return false;
            }
         }

         content.put(label, new Section(value, color));
      }

      return true;
   }

   private final boolean setContent(String contentString)
   {
      content.clear();
      datasetModified = true;

      StringTokenizer lineTokenizer = new StringTokenizer(contentString, "\n");
      while (lineTokenizer.hasMoreTokens())
         if (!parseSection(lineTokenizer.nextToken()))
            return false;

      return true;
   }

   private final String getContent()
   {
      String contentString = "";

      Iterator it = content.entrySet().iterator();
      while (it.hasNext())
      {
         Map.Entry e = (Map.Entry)it.next();
         String label = (String)e.getKey();
         Section section = (Section)e.getValue();

         contentString = contentString.concat(
              label + delim + formatValue(section.getValue()));

         Color sectionColor = section.getColor();
         if (sectionColor != null)
            contentString = contentString.concat(
               delim + "#" + HexTool.int2hexString(sectionColor.getRGB(), 3));

         contentString = contentString.concat("\n");
      }

      return contentString;
   }

   // ------------------------------------------------------------------------
   // get section[], set section/add
   // ------------------------------------------------------------------------

   private static final String SECTION_ATTR_NAME = "section";
   private static final String ADD_ATTR_NAME = "add";

   private final boolean setSection(String sectionString)
   {
      if (parseSection(sectionString))
      {
         datasetModified = true;
         return true;
      }

      return false;
   }

   private final String getSection(int sectionIndex)
   {
      Iterator it = content.entrySet().iterator();
      while (it.hasNext())
      {
         Map.Entry e = (Map.Entry)it.next();
         if (sectionIndex-- == 0)
         {
            String label = (String)e.getKey();
            Section section = (Section)e.getValue();

            String sectionString =
               label + delim + formatValue(section.getValue());

            Color sectionColor = section.getColor();
            if (sectionColor != null)
               sectionString = sectionString.concat(
                  delim + "#" + HexTool.int2hexString(sectionColor.getRGB(), 3));

            return sectionString;
         }
      }

      return "";
   }

   private final String getSection(String sectionKey)
   {
      Section section = (Section)content.get(sectionKey);
      if (section == null)
         return "";

      String sectionString =
         sectionKey + delim + formatValue(section.getValue());

      Color sectionColor = section.getColor();
      if (sectionColor != null)
         sectionString = sectionString.concat(
            delim + "#" + HexTool.int2hexString(sectionColor.getRGB(), 3));

      return sectionString;
   }

   // ------------------------------------------------------------------------
   // get sections
   // ------------------------------------------------------------------------

   private static final String SECTIONS_ATTR_NAME = "sections";

   private final int getSections()
   {
      return content.size();
   }

   // ------------------------------------------------------------------------
   // attributes
   // ------------------------------------------------------------------------

   private static final int PAINT3D_ATTR_ID = 1;
   private static final int OUTLINE_ATTR_ID = 2;
   private static final int EXPLODE_ATTR_ID = 3;
   private static final int CLEAR_ATTR_ID = 4;
   private static final int CONTENT_ATTR_ID = 5;
   private static final int SECTION_ATTR_ID = 6;
   private static final int ADD_ATTR_ID = 7;
   private static final int SECTIONS_ATTR_ID = 8;

   public void onClone(Plugin clone)
   {
      super.onClone(clone);

      Pie cloneObj = (Pie)clone;
      cloneObj.setPaint3d(getPaint3d());
      cloneObj.setOutline(getOutline());
      cloneObj.setExplode(getExplode());

      cloneObj.content.clear();
      cloneObj.datasetModified = true;

      Iterator it = content.entrySet().iterator();
      while (it.hasNext())
      {
         Map.Entry e = (Map.Entry)it.next();
         String label = (String)e.getKey();
         Section section = (Section)e.getValue();

         cloneObj.content.put(label, new Section(section));
      }
   }

   public boolean onSet(String attribute, java.util.List index,
                        UnifiedData data)
   {
      Integer attrInt = (Integer)attributeList.get(attribute);
      if (attrInt != null)
      {
         int indexElements = index != null ? index.size() : 0;
         switch (attrInt.intValue())
         {
            case PAINT3D_ATTR_ID:
               if (indexElements != 0)
                  return false;
               if (data.isStringData())
                  return false;
               setPaint3d(data.getIntData());
               break;

            case OUTLINE_ATTR_ID:
               if (indexElements != 0)
                  return false;
               if (data.isStringData())
                  return false;
               setOutline(data.getIntData());
               break;

            case EXPLODE_ATTR_ID:
               if (indexElements != 0)
                  return false;
               if (!data.isStringData())
                  return false;
               setExplode(data.getStringData());
               break;

            case CLEAR_ATTR_ID:
               if (indexElements != 0)
                  return false;
               setClear();
               break;

            case CONTENT_ATTR_ID:
               if (indexElements != 0)
                  return false;
               if (!data.isStringData())
                  return false;
               return setContent(data.getStringData());

            case SECTION_ATTR_ID:
            case ADD_ATTR_ID:
               if (indexElements != 0)
                  return false;
               if (!data.isStringData())
                  return false;
               return setSection(data.getStringData());
         }

         return true;
      }

      return super.onSet(attribute, index, data);
   }

   public UnifiedData onGet(String attribute, java.util.List index)
   {
      Integer attrInt = (Integer)attributeList.get(attribute);
      if (attrInt != null)
      {
         int indexElements = index != null ? index.size() : 0;
         switch (attrInt.intValue())
         {
            case PAINT3D_ATTR_ID:
               if (indexElements != 0)
                  return null;
               return new UnifiedData(getPaint3d());

            case OUTLINE_ATTR_ID:
               if (indexElements != 0)
                  return null;
               return new UnifiedData(getOutline());

            case EXPLODE_ATTR_ID:
               if (indexElements != 0)
                  return null;
               return new UnifiedData(getExplode());

            case CONTENT_ATTR_ID:
               if (indexElements != 0)
                  return null;
               return new UnifiedData(getContent());

            case SECTION_ATTR_ID:
            {
               if (indexElements != 1)
                  return null;
               UnifiedData sectionKey = (UnifiedData)index.get(0);
               if (!sectionKey.isStringData())
                  return new UnifiedData(getSection(sectionKey.getIntData()));
               return new UnifiedData(getSection(sectionKey.getStringData()));
            }

            case SECTIONS_ATTR_ID:
               if (indexElements != 0)
                  return null;
               return new UnifiedData(getSections());
         }
      }

      return super.onGet(attribute, index);
   }

   private static HashMap attributeList = new HashMap();
   static
   {
      attributeList.put(PAINT3D_ATTR_NAME, new Integer(PAINT3D_ATTR_ID));
      attributeList.put(OUTLINE_ATTR_NAME, new Integer(OUTLINE_ATTR_ID));
      attributeList.put(EXPLODE_ATTR_NAME, new Integer(EXPLODE_ATTR_ID));
      attributeList.put(CLEAR_ATTR_NAME, new Integer(CLEAR_ATTR_ID));
      attributeList.put(CONTENT_ATTR_NAME, new Integer(CONTENT_ATTR_ID));
      attributeList.put(SECTION_ATTR_NAME, new Integer(SECTION_ATTR_ID));
      attributeList.put(ADD_ATTR_NAME, new Integer(ADD_ATTR_ID));
      attributeList.put(SECTIONS_ATTR_NAME, new Integer(SECTIONS_ATTR_ID));
   }

   // ------------------------------------------------------------------------
   // Section, represents one pie in the chart (value and color)
   // ------------------------------------------------------------------------

   private class Section
   {
      private double value;
      private Color color;

      public Section(double value, Color color)
      {
         this.value = value;
         this.color = color;
      }

      public Section(Section other)
      {
         value = other.value;
         if (other.color != null)
            color = new Color(other.color.getRGB());
         else
            color = null;
      }

      public final void setValue(double value)
      {
         this.value = value;
      }

      public final double getValue()
      {
         return value;
      }

      public final void setColor(Color color)
      {
         this.color = color;
      }

      public final Color getColor()
      {
         return color;
      }
   }
}
